"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const chokidar_1 = require("chokidar");
const fm = require("front-matter");
const marked = require("marked");
const globby = require("globby");
const util = require("./util");
const get_image_meta_1 = require("./get-image-meta");
let pageCache = {};
let imageCache = {};
let tagCache = [];
const getDirName = (pathName) => path
    .dirname(pathName)
    .split(path.sep)
    .pop();
const sortTagsByTitle = (a, b) => a.title.localeCompare(b.title);
const sortImagesByTitle = (a, b) => a.meta.title.localeCompare(b.meta.title);
const sortTagCacheAlphabetically = (t) => t.sort(sortTagsByTitle).map((tag) => (Object.assign({}, tag, { images: tag.images.sort(sortImagesByTitle) })));
const getIndexOfTagInCache = (slug, t) => t.findIndex((item) => item.slug === slug);
const getIndexOfImageInTagCache = (t, tagIndex, filePath) => t[tagIndex].images.findIndex((img) => img.filePath === filePath);
const removeFromCache = (cache) => (filePath) => {
    if (!filePath.endsWith('.md') && !filePath.endsWith('.jpg')) {
        return;
    }
    const relativeFilePath = util.stripSlashes(filePath.replace(process.cwd(), ''));
    const cacheObject = cache;
    const item = cacheObject[relativeFilePath];
    if (!Boolean(item)) {
        return;
    }
    const newCache = Object.keys(cache)
        .filter((f) => f !== relativeFilePath)
        .reduce((acc, f) => (Object.assign({ [f]: cacheObject[f] }, acc)), {});
    Object.assign(cacheObject, newCache);
};
const addImageToTagCache = (image) => image.meta.tags.reduce((acc, tag) => {
    const tagIndex = getIndexOfTagInCache(tag.slug, acc);
    if (tagIndex === -1) {
        const tagCacheItem = Object.assign({}, tag, { images: [image] });
        return acc.concat(tagCacheItem);
    }
    const imageIndex = getIndexOfImageInTagCache(acc, tagIndex, image.filePath);
    if (imageIndex === -1) {
        acc[tagIndex].images = acc[tagIndex].images.concat(image);
    }
    else {
        acc[tagIndex].images = acc[tagIndex].images
            .slice(0, imageIndex)
            .concat([image].concat(acc[tagIndex].images.slice(imageIndex + 1)));
    }
    return acc;
}, [...tagCache]);
const removeImageFromTagCache = (filePath) => {
    if (!filePath.endsWith('.jpg')) {
        return;
    }
    const relativeFilePath = util.stripSlashes(filePath.replace(process.cwd(), ''));
    tagCache = sortTagCacheAlphabetically(tagCache.reduce((acc, tag, i) => {
        const imageIndex = getIndexOfImageInTagCache(acc, i, relativeFilePath);
        if (imageIndex === -1) {
            return acc;
        }
        if (tag.images.length === 1) {
            acc[i].images = [];
            return acc;
        }
        acc[i].images = tag.images
            .slice(0, imageIndex)
            .concat(tag.images.slice(imageIndex + 1));
        return acc;
    }, [...tagCache]));
};
const getDataForImage = (albumsDir, albumName) => (file) => ({
    filePath: file,
    detailLinkProps: util.getDetailLinkProps(albumsDir, albumName, file),
    fileName: path.basename(file),
});
const getMetaFromImage = (image) => __awaiter(this, void 0, void 0, function* () {
    const strippedImagePath = util.stripSlashes(image.filePath);
    if (imageCache[strippedImagePath]) {
        return imageCache[strippedImagePath];
    }
    const meta = yield get_image_meta_1.default(image.filePath);
    const imageWithMeta = Object.assign({}, image, { meta });
    imageCache = Object.assign({}, imageCache, { [strippedImagePath]: imageWithMeta });
    tagCache = sortTagCacheAlphabetically(addImageToTagCache(imageWithMeta));
    return imageWithMeta;
});
exports.getImages = (albumsDir, albumName) => __awaiter(this, void 0, void 0, function* () {
    const pattern = path.join('static', albumsDir, albumName, '*.jpg');
    const files = yield globby(pattern);
    const images = files.map(getDataForImage(albumsDir, albumName));
    return Promise.all(images.map(getMetaFromImage));
});
exports.getPageContent = (pagePath, pathPrefix = '') => __awaiter(this, void 0, void 0, function* () {
    const strippedPagePath = util.stripSlashes(pagePath);
    if (pageCache[strippedPagePath]) {
        return pageCache[strippedPagePath];
    }
    const content = yield util.pReadFile(pagePath);
    const { attributes, body } = fm(String(content));
    const name = getDirName(pagePath);
    pageCache = Object.assign({}, pageCache, { [strippedPagePath]: {
            meta: attributes,
            path: `/${pathPrefix}${name}/`,
            body: marked(body),
            name: name === 'content' ? 'frontpage' : name,
        } });
    return pageCache[strippedPagePath];
});
const getAlbums = (albumsDir) => (albumFile) => __awaiter(this, void 0, void 0, function* () {
    const albumName = getDirName(albumFile);
    const [content, images] = yield Promise.all([
        exports.getPageContent(albumFile, `${albumsDir}/`),
        exports.getImages(albumsDir, albumName),
    ]);
    return { content, images };
});
const getPages = (pagesGlob, albumsGlob, contentDir) => __awaiter(this, void 0, void 0, function* () {
    const files = yield globby(pagesGlob, { ignore: albumsGlob });
    return Promise.all(files.map((file) => __awaiter(this, void 0, void 0, function* () { return exports.getPageContent(file); })));
});
exports.initCachePurger = (config) => {
    const { pages } = util.getGlobPatterns(config);
    const images = path.join('static', config.albumsDir, '**/*.jpg');
    const pagesMonitor = chokidar_1.watch(pages);
    const imagesMonitor = chokidar_1.watch(images);
    pagesMonitor.on('change', removeFromCache(pageCache));
    pagesMonitor.on('unlink', removeFromCache(pageCache));
    imagesMonitor.on('change', removeFromCache(imageCache));
    imagesMonitor.on('unlink', removeFromCache(imageCache));
    imagesMonitor.on('change', removeImageFromTagCache);
    imagesMonitor.on('unlink', removeImageFromTagCache);
};
exports.getDataforTag = (config, tagSlug) => __awaiter(this, void 0, void 0, function* () {
    const tagIndex = getIndexOfTagInCache(tagSlug, tagCache);
    if (tagIndex === -1) {
        return;
    }
    return tagCache[tagIndex];
});
exports.default = (config) => __awaiter(this, void 0, void 0, function* () {
    const patterns = util.getGlobPatterns(config);
    const albumFiles = yield globby(patterns.albums);
    const albums = yield Promise.all(albumFiles.map(getAlbums(config.albumsDir)));
    const pages = yield getPages(patterns.pages, patterns.albums, config.contentDir);
    const tags = tagCache;
    return { albums, pages, tags };
});
